/* $Id: match.c,v 1.2 2000/03/15 03:39:43 ajapted Exp $
******************************************************************************

   LibGFFD - rule matching code

   Copyright (C) 2000  Andrew Apted   [andrew.apted@ggi-project.org]
   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#define GD_DEBUG_PKG    "LibGFFD"
#define GD_DEBUG_LEVEL	0

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "internal.h"


/* score table.  Lookup is [name_score][magic_score] */

static const int score_table[4][4] =
{
  {  0,  20,  50, 100 },
  { 10,  30,  70, 120 },
  { 40,  60,  80, 140 },
  { 90, 110, 130, 150 }
};


#define MATCH_SCORE(match)  \
    (score_table[(match).name_score][(match).magic_score])


static int matchOnePart(const gffd_part_t *part,
    const unsigned char *data, int data_length)
{
  unsigned char *source = part->data;
  
  int len = part->length;

  if ((part->offset + len) > data_length)
  {
    GD_DEBUG3("Not enough data too match part.\n");
    return 0;
  }

  data += part->offset;

  for (; len > 0; len--, data++, source += 2)
  {
    if ((data[0] & source[0]) != source[1])
      return 0;
  }

  /* success */
  return 1;
}

static int matchMagicParts(const gffd_magic_t *magic, 
    const unsigned char *data, int data_length)
{
  gffd_part_t *part;

  for (part = (gffd_part_t *) magic->parts.head; part != NULL;
       part = (gffd_part_t *) part->n.succ)
  {
    if (! matchOnePart(part, data, data_length))
      return 0;
  }

  /* success */
  return 1;
}

static void setupUnsortedMatches(gffd_context_t ctx)
{
  int i = 0;
  d_list_t * list = _gffdGetRules();
  gffd_rule_t *cur;

  GD_ASSERT(! ctx->matches);

  ctx->count = listCountNodes(list);
  ctx->matches = (gffd_match_t *) calloc(ctx->count, sizeof(gffd_match_t));

  for (cur = (gffd_rule_t *) list->head; cur != NULL; 
       cur = (gffd_rule_t *) cur->n.succ)
  {
    GD_ASSERT(i < ctx->count);

    ctx->matches[i].rule = cur;
    i++;
  }

  GD_ASSERT(i == ctx->count);

  ctx->full_set = 0;
}

static void bubbleDeadMatches(gffd_context_t ctx)
{
  int pos;
  gffd_match_t *M = ctx->matches;
  gffd_match_t temp;

  GD_ASSERT(M);

  pos = 0;

  while (pos+1 < ctx->count)
  {
    if (! M[pos].rule && M[pos+1].rule)
    {
      /* need to swap 'em */

      temp = M[pos];
      M[pos] = M[pos+1];
      M[pos+1] = temp;

      /* bubble down */

      if (pos > 0)
        pos--;

      continue;
    }
    else
    {
      /* bubble up */

      pos++;
    }
  }

  /* determine new count */

  for (pos = 0; pos < ctx->count && M[pos].rule; pos++)
  { }
  
  ctx->count = pos;
}

static void bubbleScoringMatches(gffd_context_t ctx)
{
  int pos;
  gffd_match_t *M = ctx->matches;
  gffd_match_t temp;

  GD_ASSERT(M);

  pos = 0;

  while (pos+1 < ctx->count)
  {
    if (MATCH_SCORE(M[pos]) < MATCH_SCORE(M[pos+1]))
    {
      /* need to swap 'em */

      temp = M[pos];
      M[pos] = M[pos+1];
      M[pos+1] = temp;

      /* bubble down */

      if (pos > 0)
        pos--;

      continue;
    }
    else
    {
      /* bubble up */

      pos++;
    }
  }
}

static void normaliseMatches(gffd_context_t ctx)
{
  if (ctx->count == 0)
    return;

  bubbleDeadMatches(ctx);

  if (! ctx->done_names && ! ctx->done_magic)
    return;
  
  bubbleScoringMatches(ctx);
}


/* ----- Comparison operations --------------------------- */


/* compares the argument name with the name(s) in the rule, returning
 * the best score, or 0 if there were no matches.
 */

static int compareName(const gffd_rule_t *rule, const char *arg)
{
  gffd_name_t *cur;

  int len = strlen(arg);
  int best = 0;

  for (cur = (gffd_name_t *) rule->names.head; cur != NULL;
       cur = (gffd_name_t *) cur->n.succ)
  {
    /* !!! FIXME: handle better wildcards than this */

    if (cur->name[0] == '*')
    {
      int suf_len = strlen(cur->name + 1);

      if (len > suf_len && strcmp(arg+(len-suf_len), cur->name+1) == 0)
        best = MAX(best, cur->score);
      
      continue;
    }

    /* exact comparison */

    if (strcmp(arg, cur->name) == 0)
      best = MAX(best, cur->score);
  }

  return best;
}


/* compares the argument filedata with the magic(s) in the rule,
 * returning the best score, or 0 if there were no matches.
 */

static int compareMagic(const gffd_rule_t *rule,
    const unsigned char *data, int length)
{
  gffd_magic_t *magic;

  int best = 0;

  for (magic = (gffd_magic_t *) rule->magics.head; magic != NULL;
       magic = (gffd_magic_t *) magic->n.succ)
  {
    if (matchMagicParts(magic, data, length))
      best = MAX(best, magic->score);
  }

  return best;
}


/* compares the argument MIME type with the one in the rule, returning
 * 1 for a match, otherwise 0.  The argument can be a partial MIME
 * type with no / to match a top level type (like "image"), or a full
 * MIME type (like "image/jpeg").  Checking is done case sensitively.
 */

static int compareMIME(const gffd_rule_t *rule, const char *arg)
{
  char *pos = strchr(arg, '/');

  /* fail if the rule has no MIME type */
  if (! rule->MIME)
    return 0;
   
  if (pos)
    return (strcmp(rule->MIME, arg) == 0);
  
  pos = strchr(rule->MIME, '/');

  GD_ASSERT(pos);
  GD_ASSERT(pos > rule->MIME);
  
  return (strlen(arg) == pos - rule->MIME) &&
         (strncmp(rule->MIME, arg, pos - rule->MIME) == 0);
}


/* ----- Matching operations --------------------------- */


int gffdMatchName(gffd_context_t ctx, const char *name)
{
  int i;

  if (! ctx->begun)
    return GFFD_E_BAD_PLACE;
  
  if (ctx->full_set)
    setupUnsortedMatches(ctx);
  
  for (i=0; i < ctx->count; i++)
  {
    gffd_match_t *match = &ctx->matches[i];

    match->name_score = compareName(match->rule, name);
  }

  ctx->done_names = 1;
  normaliseMatches(ctx);

  return 0;
}

int gffdMatchData(gffd_context_t ctx, const unsigned char *data,
    int length)
{
  int i;

  if (! ctx->begun)
    return GFFD_E_BAD_PLACE;
  
  if (ctx->full_set)
    setupUnsortedMatches(ctx);
  
  for (i=0; i < ctx->count; i++)
  {
    gffd_match_t *match = &ctx->matches[i];

    match->magic_score = compareMagic(match->rule, data, length);
  }

  ctx->done_magic = 1;
  normaliseMatches(ctx);

  return 0;
}

int gffdPruneScore(gffd_context_t ctx, int score)
{
  int i;

  if (! ctx->begun)
    return GFFD_E_BAD_PLACE;
  
  if (ctx->full_set)
    setupUnsortedMatches(ctx);
  
  for (i=0; i < ctx->count; i++)
  {
    gffd_match_t *match = &ctx->matches[i];

    /* remove entries with a too low score */
    if (MATCH_SCORE(*match) <= score)
      match->rule = NULL;
  }

  normaliseMatches(ctx);

  return 0;
}

int gffdPruneClass(gffd_context_t ctx, const char *class)
{
  int i;

  if (! ctx->begun)
    return GFFD_E_BAD_PLACE;
  
  if (ctx->full_set)
    setupUnsortedMatches(ctx);
  
  for (i=0; i < ctx->count; i++)
  {
    gffd_match_t *match = &ctx->matches[i];

    /* remove entries that do not match */
    if (strcmp(match->rule->class, class) != 0)
      match->rule = NULL;
  }

  normaliseMatches(ctx);

  return 0;
}

int gffdPruneType(gffd_context_t ctx, const char *type)
{
  int i;

  if (! ctx->begun)
    return GFFD_E_BAD_PLACE;
  
  if (ctx->full_set)
    setupUnsortedMatches(ctx);
  
  for (i=0; i < ctx->count; i++)
  {
    gffd_match_t *match = &ctx->matches[i];

    /* remove entries that do not match */
    if (strcmp(match->rule->type, type) != 0)
      match->rule = NULL;
  }

  normaliseMatches(ctx);

  return 0;
}

int gffdPruneMIME(gffd_context_t ctx, const char *MIME)
{
  int i;

  if (! ctx->begun)
    return GFFD_E_BAD_PLACE;
  
  if (ctx->full_set)
    setupUnsortedMatches(ctx);
  
  /* !!! FIXME: make argument lower case */

  for (i=0; i < ctx->count; i++)
  {
    gffd_match_t *match = &ctx->matches[i];

    /* remove entries that do not match */
    if (! compareMIME(match->rule, MIME))
      match->rule = NULL;
  }

  normaliseMatches(ctx);

  return 0;
}


/* ----- Querying operations ----------------------------- */


int gffdGetCount(gffd_context_t ctx)
{
  if (! ctx->begun)
    return GFFD_E_BAD_PLACE;
  
  if (ctx->full_set)
    setupUnsortedMatches(ctx);
  
  GD_ASSERT(! ctx->full_set);

  return ctx->count;
}

int gffdGetMatchScore(gffd_context_t ctx, int index)
{
  if (! ctx->begun || ctx->full_set)
    return GFFD_E_BAD_PLACE;
  
  if (index < 0 || index >= ctx->count)
    return GFFD_E_INVALID;
  
  GD_ASSERT(ctx->matches);
  GD_ASSERT(ctx->matches[index].rule);

  return MATCH_SCORE(ctx->matches[index]);
}

const char *gffdGetMatchClass(gffd_context_t ctx, int index)
{
  if (! ctx->begun || ctx->full_set)
    return NULL;  /* GFFD_E_BAD_PLACE */
  
  if (index < 0 || index >= ctx->count)
    return NULL;  /* GFFD_E_INVALID */

  GD_ASSERT(ctx->matches);
  GD_ASSERT(ctx->matches[index].rule);

  return ctx->matches[index].rule->class;
}

const char *gffdGetMatchType(gffd_context_t ctx, int index)
{
  if (! ctx->begun || ctx->full_set)
    return NULL;  /* GFFD_E_BAD_PLACE */
  
  if (index < 0 || index >= ctx->count)
    return NULL;  /* GFFD_E_INVALID */

  GD_ASSERT(ctx->matches);
  GD_ASSERT(ctx->matches[index].rule);

  return ctx->matches[index].rule->type;
}

const char *gffdGetMatchMIME(gffd_context_t ctx, int index)
{
  if (! ctx->begun || ctx->full_set)
    return NULL;  /* GFFD_E_BAD_PLACE */
  
  if (index < 0 || index >= ctx->count)
    return NULL;  /* GFFD_E_INVALID */

  GD_ASSERT(ctx->matches);
  GD_ASSERT(ctx->matches[index].rule);

  return ctx->matches[index].rule->MIME;
}

